export const DIRECTIVE_MAP = {
	'=': 'equals',
	'!=': 'not-equals',
	'>=': 'greater-than-equal',
	'<=': 'less-than-equal',
	'>': 'greater-than',
	'<': 'less-than',
};

export const DATA_SOURCE_QUERY_OPERATORS = Object.keys(DIRECTIVE_MAP) as (keyof typeof DIRECTIVE_MAP)[];

export type DataSourceQueryOperator = typeof DATA_SOURCE_QUERY_OPERATORS[number];

export interface DataSourceQueryExpression {
	lhs: string;
	operator: DataSourceQueryOperator;
	rhs: string | number | boolean;
}

export interface DataSourceQueryOrGrouping {
	type: 'or';
	expressions: DataSourceQueryExpression[];
}

export interface DataSourceQueryAndGrouping {
	type: 'and';
	expressions: DataSourceQueryOrGrouping[];
}

interface ProcessEntryBase {
	type: string;
}

export interface ProcessEntryString extends ProcessEntryBase {
	type: 'string';
	startsWith?: boolean;
	endsWith?: boolean;
	caseInsensitive?: boolean;
	includes?: boolean;
}

export interface ProcessEntryNumber extends ProcessEntryBase {
	type: 'number';
}

export interface ProcessEntryBoolean extends ProcessEntryBase {
	type: 'boolean';
	truthyStrings?: string[];
}

export type ProcessEntry = ProcessEntryString | ProcessEntryNumber | ProcessEntryBoolean;

interface ConvertToDataSourceQueryCollectionOptions {
	processEntries: Record<string, ProcessEntry>;
	//coerceValues: CoerceValues;
	//stringSearch: StringSearch;
}

const normalizeRhs = (rhs: string, coerceValues?: ProcessEntry) => {
	if (coerceValues?.type === 'number') {
		return Number(rhs);
	}

	if (coerceValues?.type === 'boolean') {
		const truthyStrings = [
			...(coerceValues.truthyStrings ?? []).map((s) => s.trim().toLowerCase()),
			'true',
		];

		return truthyStrings.includes(rhs);
	}

	return rhs?.toString() ?? '';
	// this will be sent to the data source, e.g. the SQL query
	// we can also make this function act as a "sanitizer"
}

export const convertToDataSourceQueryCollection = (
	q: URLSearchParams,
	options = {} as ConvertToDataSourceQueryCollectionOptions,
): DataSourceQueryAndGrouping => {
	return Array.from(q.entries()).reduce(
		(queries, [key, value]) => {
			const existingOr = queries.expressions
				.find((ex) => ex.expressions.some((ex2) => ex2.lhs === key)) ?? {
					type: 'or',
					expressions: [],
				};
			const existingLhs = existingOr.expressions.find((ex) => ex.lhs === key);
			const rhs = normalizeRhs(value, options.processEntries?.[key]);

			if (typeof existingLhs === 'undefined') {
				return {
					...queries,
					expressions: [
						...queries.expressions,
						{
							type: 'or',
							expressions: [
								{
									lhs: key,
									operator: '=',
									rhs,
								},
							],
						},
					],
				};
			}

			return {
				...queries,
				expressions: queries.expressions.map((ex) => (
					ex.expressions.some((ex2) => ex2.lhs !== key)
						? ex
						: {
							...existingOr,
							expressions: [
								...(existingOr.expressions ?? []),
								{
									lhs: key,
									operator: '=',
									rhs,
								},
							],
						}
				)),
			};
		},
		{
			type: 'and',
			expressions: [],
		} as DataSourceQueryAndGrouping
	)

	// all the queries are to be treated as "AND", as suggested by the & character to separate the query param entries
};

// TODO add more operators
